part of jaguar.throttle;

/// Interceptor to rate-limit routes
class Throttler extends FullInterceptor {
  /// Maximum number of requests allowed for the given [interval]
  final Rate quota;

  /// Identification maker. Creates an identification for the given request
  final ThrottleIdMaker idMaker;

  final Cache store;

  /// Provides an opportunity to create [Response] when rate-limit exceeds
  final RouteHandlerFunc failResponse;

  Throttler(this.quota,
      {ThrottleIdMaker idMaker, this.failResponse, Cache store})
      : store = store ?? defaultThrottleCache,
        idMaker = idMaker ?? throttleIdByIp;

  factory Throttler.perSec(int quota,
          {ThrottleIdMaker idMaker,
          RouteHandlerFunc failResponse,
          Cache store}) =>
      new Throttler(perSec(quota),
          idMaker: idMaker, failResponse: failResponse, store: store);

  factory Throttler.perMin(int quota,
          {ThrottleIdMaker idMaker,
          RouteHandlerFunc failResponse,
          Cache store}) =>
      new Throttler(perMin(quota),
          idMaker: idMaker, failResponse: failResponse, store: store);

  factory Throttler.perHour(int quota,
          {ThrottleIdMaker idMaker,
          RouteHandlerFunc failResponse,
          Cache store}) =>
      new Throttler(perHour(quota),
          idMaker: idMaker, failResponse: failResponse, store: store);

  factory Throttler.perDay(int quota,
          {ThrottleIdMaker idMaker,
          RouteHandlerFunc failResponse,
          Cache store}) =>
      new Throttler(perDay(quota),
          idMaker: idMaker, failResponse: failResponse, store: store);

  ThrottleState _state;

  ThrottleState get output => _state;

  Future before(Context ctx) async {
    final String throttleId = await idMaker(ctx);
    final ThrottleData count = await _visit(throttleId, quota.interval);

    _state = new ThrottleState.make(quota.count, count, quota.interval);

    if (count == null) return _state;

    if (count.count > quota.count) {
      final Response resp = failResponse != null
          ? await failResponse(ctx)
          : new Response('Limit exceeded', statusCode: 429);

      throw _state.setHeaders(resp);
    }

    ctx.addInterceptor(Throttler, id, this);
    return _state;
  }

  Response after(Context ctx, Response incoming) {
    _state.setHeaders(incoming);
    return incoming;
  }

  Future<ThrottleData> _visit(String identifier, Duration interval) async {
    ThrottleData ret = new ThrottleData.now(1);

    ThrottleData data;

    try {
      data = await store.read(identifier);
    } catch (e) {
      if (e != cacheMiss) rethrow;
    }

    final now = new DateTime.now();

    if (data == null) {
      ret = new ThrottleData.now(1);
    } else if (now.difference(data.time()) < interval) {
      ret = data.inc(data.milliseconds);
    } else {
      ret = new ThrottleData.now(1);
    }

    await store.upsert(identifier, ret);
    return ret;
  }
}
